import numpy as np
import matplotlib.pyplot as plt
import nibabel as nib
from tqdm.contrib import tzip
from scipy.optimize import minimize
from scipy.linalg import block_diag
import warnings
from tqdm import tqdm

class SignalDSC():
    def __init__(self, dsc_path, brain_mask_path, aif_mask_path=None, baseline=8, offset=0, r2=1, TE=0.03, nobar=True, plot=False):
        self.baseline = baseline
        self.r2 = r2
        self.TE = TE
        self.kH = 0.1369 * (0.55 / 0.75) # Constant grouping tissular parameters for absolute quantification
        self.rho_Voi = 0.0104  # Apparent brain density [100g/ml]
        self.hematocrit = self.kH / self.rho_Voi
        self.offset = offset
        self.plot = plot

        functions = [
            (self.load_dsc, dsc_path, "Loading DSC data"),
            (self.load_brain_mask, brain_mask_path, "Loading brain mask"),
            (self.load_aif_mask, aif_mask_path, "Loading AIF mask"),
            (self.load_signal_in_brain_mask, None, "Extracting DSC signals among the brain mask")
        ]

        if nobar:
            self.load_dsc(dsc_path)
            self.load_brain_mask(brain_mask_path)
            self.load_aif_mask(aif_mask_path)
            self.load_signal_in_brain_mask()
        else:
            # Create a tqdm progress bar
            with tqdm(total=len(functions), desc="Processing DSC-MRI Perfusion data") as pbar:
                for func, arg, desc in functions:
                    pbar.set_description(f"Processing DSC-MRI Perfusion data - {desc}")
                    if arg is not None:
                        func(arg) 
                    else:
                        func()
                    pbar.update(1) 
 
    def load_dsc(self, dsc_path):
        self.dsc_path = dsc_path
        self.dsc = nib.load(dsc_path)
        self.dsc_data = self.dsc.get_fdata()
        self.size = self.dsc_data.shape[-1]

    def load_brain_mask(self, brain_mask_path):
        self.brain_mask_path = brain_mask_path
        self.brain_mask = nib.load(brain_mask_path)
        self.brain_mask_data = self.brain_mask.get_fdata().astype(bool)
        self.affine = self.brain_mask.affine

    def load_aif_mask(self, aif_mask_path):
        self.aif_mask_path = aif_mask_path
        self.aif_mask = nib.load(aif_mask_path)
        self.aif_mask_data = self.aif_mask.get_fdata().astype(bool)

        self.aif_coord = np.argwhere(self.aif_mask_data > 0)
        self.AIF = np.mean([self.dsc_data[aif[0], aif[1], aif[2], :] for aif in self.aif_coord], axis=0)[self.offset:]
        self.AIF_0 = np.mean(self.AIF[:self.baseline])
        self.Ca = - np.log(self.AIF / self.AIF_0 + 1e-9) / (self.r2 * self.TE)
        self.Ca = np.where(self.Ca < 0., 0, self.Ca)
        self.Ca_0 = np.mean(self.Ca[:self.baseline])

        self.Ca_gamma = self.gamma_variate_fitting(plot=self.plot)

    def load_signal_in_brain_mask(self):
        self.S = self.dsc_data[:,:,:,self.offset:]
        self.S_0 = np.mean(self.dsc_data[:, :, :, self.offset:self.baseline], axis=-1)

        self.S = self.S[self.brain_mask_data]
        self.S_0 = self.S_0[self.brain_mask_data]

    def signal_to_concentration(self, clip=True):
        self.C = - np.log(self.S / np.broadcast_to(self.S_0[:,np.newaxis], self.S.shape) + 1e-9) / (self.r2 * self.TE)
        self.C = np.nan_to_num(self.C, nan=0.)
        if clip: self.C = np.where(self.C < 0, 0., self.C)
    
    def concentration_to_signal(self, aif=False):
        if aif: self.S_revert = self.Ca_0 * np.exp(-self.TE * self.C)
        else: self.S_revert = self.S_0[:, np.newaxis] * np.exp(-self.TE * self.C)
        return self.S_revert

    def gamma_variate_fitting(self, plot=True):

        def gamma_variate(t, t0, r, b, C0):
                tr = t - t0
                y = C0 * (np.sign(tr) * (np.abs(tr)) ** r) * np.exp((-tr) / b)
                return np.where(t > t0, y, 0)

        def gamma_variate_loss(Parameters, Inputs):
            t0 = Parameters[0]
            r = Parameters[1]
            b = Parameters[2]
            C0 = Parameters[3]
            Time = Inputs[0]
            Intensity = Inputs[1]
            Predictions = gamma_variate(Time, t0, r, b, C0)
            return ((Predictions - Intensity) ** 2).mean()

        size = self.Ca.shape[0]
        time_var = np.array([i for i in range(size)])

        dt = 1
        first_derivative = np.diff(self.Ca) / dt

        # Define a threshold
        threshold = np.mean(first_derivative) + 3 * np.std(first_derivative)  # 3-sigma rule

        # Detect the bolus arrival time (BAT)
        try:
            bat_index = np.where(first_derivative > threshold)[0][0]
        except: 
            bat_index = self.baseline
            
        bat_time = time_var[bat_index]

        print(f"BAT: {bat_time}")

        time_var = time_var[:size]
        aif4fitt = self.Ca[:size]

        InputList = [time_var, aif4fitt]

        Optmizer = minimize(
            gamma_variate_loss, [bat_index,1,1,1], InputList,
            # method='nelder-mead',
            # method='SLSQP',
            method='L-BFGS-B',
            options={'xtol': 1e-9, 'disp': False, 'maxiter': 100000}
            )
        t0 = Optmizer.x[0]
        r = Optmizer.x[1]
        b = Optmizer.x[2]
        C0 = Optmizer.x[3]

        fitted_aif = gamma_variate(time_var, bat_index, r, b, C0)

        if plot:
            plt.figure()
            plt.plot(time_var, aif4fitt, 'o')
            plt.plot(time_var, fitted_aif, '-')
            plt.legend(['AIF', 'GammaVariateAIF'])
            plt.xlabel("Time")
            plt.ylabel(r"C$_a$ (t)")
            plt.title("Gamma Variate function fitting output in the time interval [0, {}]".format(size))
            plt.show()
        
        time_var = np.array([i for i in range(size)])
        fitted_aif = gamma_variate(time_var, t0, r, b, C0)

        if plot:
            plt.figure()
            plt.plot(time_var, fitted_aif, '-')
            plt.plot(time_var, self.Ca, 'o', markersize=2)
            plt.legend(['AIF', 'GammaVariateAIF'])
            plt.xlabel("Time")
            plt.ylabel(r"C$_a$ (t)")
            plt.title("Gamma Variate function fitting output in the entire curve")
            plt.show()

        # Once again, but for the AIF fitted
        dt = time_var[1] - time_var[0]
        first_derivative = np.diff(fitted_aif) / dt

        # Define a threshold
        threshold = np.mean(first_derivative) + 3 * np.std(first_derivative)  # 3-sigma rule

        # Detect the bolus arrival time (BAT)
        try:
            bat_index = np.where(first_derivative > threshold)[0][0]
        except: 
            bat_index = self.baseline
        bat_time = time_var[bat_index]

        # Find the peak time (t_peak)
        peak_index = np.argmax(fitted_aif[bat_index:]) + bat_index
        peak_time = time_var[peak_index]

        # Calculate TTP
        ttp = peak_time - bat_time
        cbf = (fitted_aif[peak_index] - fitted_aif[bat_index]) / (peak_time - bat_time)

        # Calculate CBV
        cbv = np.trapz(fitted_aif[bat_index:size], time_var[bat_index:size])

        # Plotting the signal and the detected BAT
        if plot:
            plt.figure(figsize=(10, 5))
            plt.plot(time_var, fitted_aif, label='Signal')
            plt.axvline(bat_time, color='r', linestyle='--', label=f'BAT: {bat_time:.2f} s')
            plt.axvline(peak_time, color='g', linestyle='--', label=f'Peak Time: {peak_time:.2f} s, TTP: {ttp:.2f} s')
            plt.fill_between(time_var[bat_index:], fitted_aif[bat_index:], alpha=0.5, color='orange', label=f'CBV: {cbv:.2f} ua')
            plt.plot([], [], alpha=0, label=f'CBF: {cbf:.2f}')
            plt.plot([], [], alpha=0, label=f'MTT: {(cbv/cbf):.2f}')
            plt.xlabel('Time (seconds)')
            plt.ylabel('Signal Intensity')
            plt.title(f'Bolus Arrival Time Detection in DSC Perfusion MRI')
            plt.legend()
            plt.show()

        return fitted_aif
    
    def solve_CBV(self, clip=True):
        CBV_ = np.trapz(self.C, dx=1, axis=-1) / np.trapz(self.Ca)
        CBV_ = self.hematocrit * CBV_
        self.CBV = np.zeros_like(self.brain_mask_data).astype(np.float32)
        self.CBV[self.brain_mask_data] = CBV_
        if clip: self.CBV = np.where(self.CBV < 0, 0, self.CBV)
        return self.save_image(self.CBV, "pf_CBV")

    def solve_CBF(self):
        self.CBF = np.max(self.TRF, axis=-1)
        self.CBF = self.hematocrit * self.CBF
        return self.save_image(self.CBF * 60, "pf_CBF")

    def solve_MTT(self):
        self.MTT = self.CBV / (self.CBF + 1e-9)
        return self.save_image(self.MTT, "pf_MTT")

    def solve_TTP(self):
        TTP_ = np.array([np.argmax(signal) for signal in self.C]).astype(np.float32)
        self.TTP = np.zeros_like(self.brain_mask_data).astype(np.float32)
        self.TTP[self.brain_mask_data] = TTP_
        return self.save_image(self.TTP, "pf_TTP")

    def solve_Tmax(self, threshold=6.):
        Tmax_ = np.array([np.argmax(signal) for signal in self.TRF_]).astype(np.float32)
        self.Tmax = np.zeros_like(self.brain_mask_data).astype(np.float32)
        self.Tmax[self.brain_mask_data] = Tmax_
        self.save_image(self.Tmax, "pf_Tmax")
        Tmax_ = np.where(self.Tmax < threshold, 0, 1)
        return self.save_image(Tmax_, f"pf_Tmax_{str(threshold)}")

    def solve_signal_recovery(self, S_pre, S_post, initial_frame=0):
        self.S_min = np.argmin(np.mean(self.dsc_data, axis=(0,1,2)))
        self.S_pre = [i - initial_frame for i in [initial_frame, S_pre]]
        self.S_post = [i - initial_frame for i in [S_post, self.S.shape[-1]]]

        shifted_S_pre = [i - initial_frame for i in self.S_pre]
        shifted_S_post = [i - initial_frame for i in self.S_post]
        shifted_S_min = self.S_min - initial_frame

        SR_num = np.mean(self.dsc_data[:,:,:,shifted_S_post[0]:shifted_S_post[1]], axis=-1) - np.mean(self.dsc_data[:,:,:,shifted_S_pre[0]:shifted_S_pre[1]], axis=-1)
        SR_denom = np.mean(self.dsc_data[:,:,:,shifted_S_pre[0]:shifted_S_pre[1]], axis=-1)
        self.SR = 100 * SR_num / (SR_denom + 1e-9)

        PSR_num = np.mean(self.dsc_data[:,:,:,shifted_S_post[0]:shifted_S_post[1]], axis=-1) - self.dsc_data[:,:,:,shifted_S_min]
        PSR_denom = np.mean(self.dsc_data[:,:,:,shifted_S_pre[0]:shifted_S_pre[1]], axis=-1) - self.dsc_data[:,:,:,shifted_S_min]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            self.PSR = 100 * PSR_num / (PSR_denom + 1e-9)

        self.SR *= self.brain_mask_data
        self.PSR *= self.brain_mask_data
        self.save_image(self.SR, "pf_SR")
        self.save_image(self.PSR, "pf_PSR")

    def solve_deconvolution(self, mode="svd"):
        def _svd(aif):
            temp = np.array([np.roll(aif, i) for i in range(aif.shape[0])]).T
            Lambda = np.tril(temp)
            u, s, vh = np.linalg.svd(Lambda)
            return u, s, vh
        
        def _osvd(aif, oi_threshold=0.1):
            # Create the block-circulant matrix
            n = len(aif)
            aif_padded = np.concatenate([aif, np.zeros(n - 1)])
            Lambda = block_diag(*[np.roll(aif_padded, i)[:n] for i in range(n)])
            
            # Perform SVD
            u, s, vh = np.linalg.svd(Lambda)
            
            # Calculate the oscillation index (OI)
            diags = np.array([np.diag(vh, k) for k in range(-n + 1, n)])
            oscillations = np.sum(np.abs(np.diff(diags, axis=1)), axis=1)
            total_variation = np.sum(np.abs(diags), axis=1)
            oi = oscillations / (total_variation + np.finfo(float).eps)
            
            # Determine the filter based on OI threshold
            f_j = oi < oi_threshold
            return u, s, vh, f_j
        
        def _deconv(c, U, S, Vh, f_j):
            invS = S ** -1
            invSigma = np.where(invS == np.inf, 0, invS) 
            inv_AIF = np.dot(Vh.T, np.dot(np.diag(f_j * invSigma), U.T))
            k = np.dot(inv_AIF, c)
            return k
        
        self.shift = np.argmax(self.Ca_gamma > 0.01 * np.max(self.Ca_gamma))
        print(f"Shifted {self.shift} frames of the AIF")

        aif = self.Ca_gamma[self.shift:]
        C = np.zeros_like(self.dsc_data)[:,:,:,self.offset:]
        C[self.brain_mask_data] = self.C
        C = C[:,:,:,self.shift:]

        if mode == "svd":
            u, s, v = _svd(aif)
            f_j = np.ones_like(s)
        elif mode == "tsvd":
            u, s, v = _svd(aif)
            f_j = s > (0.2 * s[0])
        elif mode == "tikhonov":
            u, s, v = _svd(aif)
            f_j = (s ** 2) / (s ** 2 + (0.2 * s[0]) ** 2)
        elif mode == "osvd":
            u, s, v, f_j = _osvd(aif)
        else: 
            raise Exception(f"Mode must be chosen between 'svd', 'tsvd', 'tikhonov', or 'osvd', but found {mode}.")

        self.TRF = np.zeros_like(C)
        non_zero_indices = np.nonzero(self.brain_mask_data)
        print(f"Performing {mode} deconvolution...")
        for (i, j, k) in tzip(*non_zero_indices):
            self.TRF[i, j, k, :] = _deconv(C[i, j, k, :], u, s, v, f_j)

        self.TRF_ = self.TRF[self.brain_mask_data]

    def save_image(self, volume, suffix):
        output_path = self.dsc_path.replace(".nii", f"_{suffix}.nii")
        nib.save(
            nib.Nifti1Image(
                volume, self.affine
            ),
            output_path
        )
        return output_path

def single_deconvolution(C, Ca, mode="svd"):
    temp = np.array([np.roll(Ca, i) for i in range(Ca.shape[0])]).T
    Lambda = np.tril(temp)
    u, s, vh = np.linalg.svd(Lambda)
    if mode == "svd":
        f_j = np.ones_like(s)
    elif mode == "tsvd":
        f_j = s > (0.2 * s[0])
    elif mode == "tikhonov":
        f_j = (s ** 2) / (s ** 2 + (0.2 * s[0]) ** 2)
    else: 
        raise Exception(f"Mode parameter must be chosen between 'svd', 'tsvd' or 'tikhonov', but found {mode}.")
    invS = s ** -1
    invSigma = np.where(invS == np.inf, 0, invS) 
    inv_Ca = np.dot(vh.T, np.dot(np.diag(f_j * invSigma), u.T))
    k = np.dot(inv_Ca, C)
    return k



import os
import SimpleITK as sitk
import torchio as tio
import nibabel as nib
import numpy as np
from skimage import morphology
import matplotlib.pyplot as plt
from .brain_extraction import run_hd_bet
from dipy.segment.tissue import TissueClassifierHMRF
import itk
from numba import jit
import torchio as tio
import torch
from .models.mpunet import MPUnetPP2CBND

def n4_bias_field_correction(input_path, mask=False, force=True, suffix="corrected"):
    """ Correct the bias field correction.

    Parameters
    ----------
    input_path : str
        Nifti file path

    suffix : str
        Nifti file suffixe for the new generated image

    mask_path : str
        Nifti file path containing the mask of the bias field

    """
    print(f"INFO - Starting bias field correction for\n\t{input_path :}")
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    mask_path = output_path.replace(".nii.gz", "_mask.nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Bias field correction already done for\n\t{input_path :}")
        if mask is True: 
            return output_path, mask_path
        else:
            return output_path

    nifti = sitk.ReadImage(input_path, sitk.sitkFloat32)
    mask_otsu = sitk.OtsuThreshold(nifti,0,1,200)
    corrector = sitk.N4BiasFieldCorrectionImageFilter()

    output = corrector.Execute(nifti, mask_otsu)
    print(f"INFO - Saving generated image at\n\t{output_path :}")

    sitk.WriteImage(output, output_path)

    if mask is True: 
        print(f"INFO - Saving generated mask at\n\t{mask_path :}")
        log_bias_field = corrector.GetLogBiasFieldAsImage(nifti)
        sitk.WriteImage(log_bias_field, mask_path)
        return output_path, mask_path
    else:
        return output_path

def registration(atlas_path, input_path, matrix=True, force=True, suffix="flirt"):
    """ Register a nifti image to an atlas (MNI152).

    Parameters
    ----------
    atlas_path : str
        Nifti file path which is the atlas 

    input_path : str
        Nifti file path to be registered

    suffix : str
        Nifti file suffixe for the new generated image

    """
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    matrix_path = input_path.replace(".nii.gz", "_transfo.tfm")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Registration already done for\n\t{input_path :}\n\t{atlas_path :}")
        if matrix == True:
            return output_path, matrix_path
        else:
            return output_path

    print(f"INFO - Starting registration for\n\t{input_path :}\n\t{atlas_path :}")
    fixed = sitk.ReadImage(atlas_path, sitk.sitkFloat32)
    moving = sitk.ReadImage(input_path, sitk.sitkFloat32)

    R = sitk.ImageRegistrationMethod()
    R.SetMetricAsMattesMutualInformation()
    R.SetOptimizerAsPowell()
    R.SetInitialTransform(sitk.TranslationTransform(fixed.GetDimension()))
    R.SetInterpolator(sitk.sitkLinear)
    R.SetOptimizerScalesFromPhysicalShift()

    outTx = R.Execute(fixed, moving)

    resampled = sitk.Resample(moving,
                                fixed,
                                outTx,
                                sitk.sitkLinear, 0.0,
                                moving.GetPixelIDValue())

    print(f"INFO - Saving generated image at\n\t{output_path :}")
    sitk.WriteImage(resampled, output_path)

    if matrix is True: 
        sitk.WriteTransform(outTx, matrix_path)
        return output_path, matrix_path
    else:
        return output_path

def apply_transformation(atlas_path, input_path, matrix_path, force=True, suffix="flirt"):
    """ Apply a transformation to an image according the transformation matrix .

    Parameters
    ----------
    atlas_path : str
        Nifti file path which is the atlas 

    input_path : str
        Nifti file path to be registered

    matrix_path : str
        Matrix path that will be use to transform the input image

    suffix : str
        Nifti file suffixe for the new generated image

    """
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Registration already done for\n\t{input_path :}\n\t{atlas_path :}")
        return output_path

    print(f"INFO - Starting registration for\n\t{input_path :}\n\t{atlas_path :}")
    fixed = sitk.ReadImage(atlas_path, sitk.sitkFloat32)
    moving = sitk.ReadImage(input_path, sitk.sitkFloat32)

    outTx = sitk.ReadTransform(matrix_path)

    resampled = sitk.Resample(moving,
                                fixed,
                                outTx,
                                sitk.sitkLinear,
                                0.0,
                                moving.GetPixelIDValue())

    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    print(f"INFO - Saving generated image at\n\t{output_path :}")
    sitk.WriteImage(resampled, output_path)

    return output_path

def resample(*values, suffix="resampled"):
    """ Resample images in the same space with a resample factor. If no resample factor is given, the first image become the fixed image and all others paths will be resampled on it

    Parameters
    ----------
    input_paths : str 
        All the image paths which need to be resampled on a fixed image or on a resampled value

    resample : int (optional)
        The resample value in which all the image will be transformed.

    
    Example
    -------
    resample(t1_path, 1) => the t1_path is resampled with 1mm voxel size
    resample(t1_path, t2_path) => t1_path is the fixed image, t2_path is the moving image
    resample(t1_path, t2_path, swi_path, 2) => all images are resampled with 2mm voxel size
    resample(t1_path) => Not working because if there is any resample value, 2 paths are mendatory
    """
    nl = "\n\t"
    temp = values
    if temp[-1] is not str:
        temp = temp[0:-1]
    print(f"INFO - Starting resampling for{nl}{nl.join(temp) :}")
    new_paths = []
    paths = []
    transform = None
    for value in values:
        if type(value) == str and os.path.exists(value):
            paths.append(value)

        elif type(value) == int or type(value) == float:
            transform = tio.Resample(value)

    if len(paths) > 0:
        if transform == None:
            if len(paths) >= 2:
                subject = tio.Subject(ref=tio.ScalarImage(paths.pop(0)))

                for path in paths:
                    subject.add_image(tio.ScalarImage(path), path)

                transform = tio.Resample("ref")
                subject = transform(subject)

                for image in subject.get_images_names():
                    if image != "ref":
                        new_path = image.replace(".nii.gz", "_" + suffix + ".nii.gz")
                        subject[image].save(new_path)
                        new_paths.append(new_path)
            else:
                raise ValueError("Need at least 2 valid paths (1 as ref, 1 as resampled) when no resampling value is given")
        else:
            subject = tio.Subject(ref=tio.ScalarImage(paths[0]))
            for path in paths:
                subject.add_image(tio.ScalarImage(path), path)

            subject = transform(subject)

            for image in subject.get_images_names():
                if image != "ref":
                    new_path = image.replace(".nii.gz", "_" + suffix + ".nii.gz")
                    subject[image].save(new_path)
                    new_paths.append(new_path)

    else:
        raise ValueError("All the given path were not valid")
    
    return new_paths

def histogram_matching(input_ref_path, input_path, suffix="hm"):
    """ Match the histogram of two images.

    Parameters
    ----------
    input_ref_path : str
        The fixed image on which the other image will be matched

    input_path : str 
        The moving image that needs to be filtered

    suffix : str
        Nifti file suffixe for the new generated image
    """
    print(f"INFO - Starting histogram matching for\n\t{input_ref_path :}\n\t{input_path :}")
    him = sitk.HistogramMatchingImageFilter()
    him.SetThresholdAtMeanIntensity(True)

    readerI = sitk.ImageFileReader()
    readerI.SetFileName(input_path)
    in_img = readerI.Execute()

    readerII = sitk.ImageFileReader()
    readerII.SetFileName(input_ref_path)
    ref = readerII.Execute()

    out_img = him.Execute(in_img,ref)

    writer = sitk.ImageFileWriter()

    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    print(f"INFO - Saving generated image at\n\t{output_path :}")

    writer.SetFileName(output_path)
    writer.Execute(out_img)

def apply_mask(input_path, mask_path, force=True, suffix="mask"):
    """ Apply a binary mask to an image sharing the same shape.

    Parameters
    ----------
    input_path : str
        The image path on which apply the mask 

    mask_path : str 
        The binary mask path 

    suffix : str
        Nifti file suffixe for the new generated image
    """
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Mask already applied for\n\t{input_path :}")
        return output_path

    print(f"INFO - Starting to apply mask for\n\t{input_path :}\n\t{mask_path :}")
    img = nib.load(input_path)
    mask = nib.load(mask_path)

    if mask.shape != img.shape:
        raise(f"ERROR - Shape of the mask does not match the main image :\n\t Image 1 : {img.shape} != Image 2 : {mask.shape}")

    final = mask.get_fdata() * img.get_fdata()

    print(f"INFO - Saving generated image at\n\t{output_path :}")

    nib.save(nib.Nifti1Image(final, img.affine), output_path)

    return output_path

def substract(input_ref_path, input_path, is_aligned=True, suffix="sub"):
    """ Match the histogram of two images.

    Parameters
    ----------
    input_ref_path : str
        The fixed image that will be substracted

    input_path : str 
        The image on which the other image will be substracted

    is_aligned : Boolean
        Boolean to check if the images are aligned. If not, it starts registration function before substraction

    suffix : str
        Nifti file suffixe for the new generated image
    """
    print(f"INFO - Starting substraction for\n\t{input_ref_path :}\n\t{input_path :}")
    ref = nib.load(input_ref_path)

    if is_aligned == False:
        input_aligned_path = registration(input_ref_path, input_path, suffix="flirt")
        img = nib.load(input_aligned_path)
    else:
        print(f"INFO - Assuming the images are already aligned. Registration has been skipped")
        img = nib.load(input_path)


    sub = ref.get_fdata() - img.get_fdata()

    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    print(f"INFO - Saving generated image at\n\t{output_path :}")

    nib.save(nib.Nifti1Image(sub, img.affine), output_path)

    return output_path

def susan(input_path, offset=3, force=True, suffix="susan"):
    """ Amplify borders by creating a susan variance map.
    Based on https://users.fmrib.ox.ac.uk/~steve/susan/susan/node2.html

    Parameters
    ----------
    input_path : str
        The image that will be used to create the susan variance map

    offset : int
        Size of the kernel

    suffix : str
        Nifti file suffixe for the new generated image
    """
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Susan variance map already computed for\n\t{input_path :}")
        return output_path

    print(f"INFO - Starting Susan variance map creation for\n\t{input_path :}")
    img = nib.load(input_path)
    img_data = img.get_fdata()

    @jit
    def susan(img_data):
        susan_data = np.zeros_like(img_data)
        for x in range(img_data.shape[0]):
            for y in range(img_data.shape[1]):
                for z in range(img_data.shape[2]):
                        values = img_data[x:x + offset, y:y + offset, z:z + offset]
                        mean = np.mean(values)
                        var = np.var(values)
                        susan_data[x,y,z] = np.abs(var - mean**2)
        return susan_data
    
    susan_data = susan(img_data)

    print(f"INFO - Saving generated image at\n\t{output_path :}")
    nib.save(nib.Nifti1Image(susan_data, img.affine), output_path)
    
    return output_path

def normalize(input_path, force=True, suffix="minmax"):
    """ Normalize image between 0 and 1.

    Parameters
    ----------
    input_path : str
        The image that will be normalized

    suffix : str
        Nifti file suffixe for the new generated image
    """
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Normalization already computed for\n\t{input_path :}")
        return output_path

    print(f"INFO - Starting normalization for\n\t{input_path :}")
    img = nib.load(input_path)
    img_data = img.get_fdata()
    norm_data = (img_data - img_data.min()) / (img_data.max() - img_data.min())
    
    print(f"INFO - Saving generated image at\n\t{output_path :}")
    nib.save(nib.Nifti1Image(norm_data, img.affine), output_path)

    return output_path

def crop(input_path, force=True, suffix="cropped"):
    """ Crop the maximum of noise in an image.

    Parameters
    ----------
    input_path : str
        The image that will bet cropped

    suffix : str
        Nifti file suffixe for the new generated image

    Returns
    -------
    cropped_idx : list if int
        List of the indices that was used to crop the input image

    """
    print(f"INFO - Starting automatic crop for\n\t{input_path :}")
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    
    mri = nib.load(input_path)
    mri_image = mri.get_fdata()
    
    z_dim, y_dim, x_dim = mri_image.shape

    # Calculate the threshold for the image intensity to determine the background
    threshold = np.mean(mri_image) + 2 * np.std(mri_image)

    # Find the maximum intensity value within a margin around the center of the image
    z_margin, y_margin, x_margin = z_dim // 4, y_dim // 4, x_dim // 4
    z_center, y_center, x_center = z_dim // 2, y_dim // 2, x_dim // 2
    center_intensity = np.max(mri_image[z_center - z_margin:z_center + z_margin,
                                        y_center - y_margin:y_center + y_margin,
                                        x_center - x_margin:x_center + x_margin])

    # Set a lower threshold to ensure that the cube contains some useful information
    lower_threshold = center_intensity * 0.75

    # Determine the starting and ending indices for each axis
    z_start = np.argmax(np.max(mri_image, axis=(1, 2)) > threshold)
    z_end = z_dim - np.argmax(np.max(mri_image[::-1, :, :], axis=(1, 2)) > threshold)
    y_start = np.argmax(np.max(mri_image, axis=(0, 2)) > threshold)
    y_end = y_dim - np.argmax(np.max(mri_image[:, ::-1, :], axis=(0, 2)) > threshold)
    x_start = np.argmax(np.max(mri_image, axis=(0, 1)) > threshold)
    x_end = x_dim - np.argmax(np.max(mri_image[:, :, ::-1], axis=(0, 1)) > threshold)

    # Make sure the cube contains some useful information
    while (mri_image[z_start:z_end, y_start:y_end, x_start:x_end].max() < lower_threshold):
        z_start += 1
        z_end -= 1
        y_start += 1
        y_end -= 1
        x_start += 1
        x_end -= 1

    cropped_idx = [z_start, z_end, y_start, y_end, x_start, x_end]
    if os.path.exists(output_path) and not force:
        print(f"INFO - Automatic crop already done for\n\t{input_path :}")
        return output_path, cropped_idx
    
    # Print the starting and ending indices for each axis
    # print(f"z_start = {z_start}, z_end = {z_end}")
    # print(f"y_start = {y_start}, y_end = {y_end}")
    # print(f"x_start = {x_start}, x_end = {x_end}")

    # Crop the MRI image to extract the cube
    cropped_mri_image = mri_image[z_start:z_end, y_start:y_end, x_start:x_end]
    nib.save(nib.Nifti1Image(cropped_mri_image, mri.affine), output_path)

    print(f"INFO - Saving generated image at\n\t{output_path :}")
    return output_path, cropped_idx

def apply_crop(input_path, crop, force=True, suffix="cropped"):
    """ Crop the image according crop indices.

    Parameters
    ----------
    input_path : str
        The image that will get the new affine from the reference image

    crop : list
        List of 6 integers to determine the cropped indices 

    suffix : str
        Nifti file suffixe for the new generated image
    """
    print(f"INFO - Starting automatic crop for\n\t{input_path :}")
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Automatic crop already done for\n\t{input_path :}")
        return output_path

    mri = nib.load(input_path)
    mri_image = mri.get_fdata()

    cropped_mri_image = mri_image[crop[0]:crop[1], crop[2]:crop[3], crop[4]:crop[5]]
    nib.save(nib.Nifti1Image(cropped_mri_image, mri.affine), output_path)

    print(f"INFO - Saving generated image at\n\t{output_path :}")
    return output_path

def copy_affine(input_ref_path, input_path, force=True, suffix="affined"):
    """ Match the histogram of two images.

    Parameters
    ----------
    input_ref_path : str
        The image used as reference

    input_path : str
        The image that will get the new affine from the reference image

    suffix : str
        Nifti file suffixe for the new generated image
    """
    print(f"INFO - Starting copying affine for\n\t{input_path :}\n\tWith the following image\n\t{input_ref_path :}")

    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Copying affine already done for\n\t{input_path :}")
        return output_path

    subject = tio.Subject(
        ref=tio.ScalarImage(input_ref_path),
        img=tio.ScalarImage(input_path)
    )
    subject = tio.CopyAffine('ref')(subject)

    for image in subject.get_images_names():
        if image == "img":
            subject[image].save(image.replace(".nii.gz", "_" + suffix + ".nii.gz"))

    print(f"INFO - Saving generated image at\n\t{output_path :}")
    return output_path

def skull_stripping(input_path, model_path, mask=False, force=True, suffix="brain"):
    """ HD-BET to extract the brain of any anatomic mri image.
    This script is a slightly modified version of the HD-BET script that can be found at the following link:
    https://github.com/MIC-DKFZ/HD-BET

    Isensee F, Schell M, Tursunova I, Brugnara G, Bonekamp D, Neuberger U, Wick A, Schlemmer HP, Heiland S, Wick W,
    Bendszus M, Maier-Hein KH, Kickingereder P. Automated brain extraction of multi-sequence MRI using artificia
    neural networks. arXiv preprint arXiv:1901.11341, 2019.
    https://doi.org/10.1002/hbm.24750

    Parameters
    ----------
    input_path : str
        The image path in which the brain will be extracted

    model_path : str
        Path of the directory containing the folds of the model. 
        The files can be downloaded from the following path: 
        https://zenodo.org/record/2540695

    mask : bool
        To get the binary mask of the brain
        
    suffix : str
        Nifti file suffixe for the new generated image

    """
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    mask_path = output_path.replace(".nii.gz", "_mask.nii.gz")

    print(f"INFO - Starting skull stripping for\n\t{input_path :}")
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Skull stripping already done for\n\t{input_path :}")
        if mask == True:
            return output_path, mask_path
        else:
            return output_path
    
    img = nib.load(input_path)
    data = np.transpose(img.get_fdata(), (2,1,0)) 
    sitk_data = sitk.GetImageFromArray(data)
    img_out, mask_out = run_hd_bet(sitk_data, model_path=model_path)
    img_out = np.transpose(img_out, (2,1,0)) 
    mask_out = np.transpose(mask_out, (2,1,0)) 

    print(f"INFO - Saving generated image at\n\t{output_path :}")
    nib.save(nib.Nifti1Image(img_out, img.affine), output_path)
    if mask == True:
        print(f"\t{mask_path :}")
        nib.save(nib.Nifti1Image(mask_out, img.affine), mask_path)
        return output_path, mask_path
    else:
        return output_path

def remove_small_object(input_path, min_size, force=False, suffix="filtered"):
    """ Remove small objects above a minimum size

    Parameters
    ----------
    input_path : str
        Nifti file path

    min_size : int
        Minimum size of objects to be removed

    suffix : str
        Nifti file suffixe for the new generated image

    """
    print(f"INFO - Starting to remove small objects of size inferior to {min_size} for\n\t{input_path :}")
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Small object removing already done for\n\t{input_path :}")
        return output_path
    
    img = nib.load(input_path)
    img_data = img.get_fdata()

    img_out = morphology.remove_small_objects(img_data.astype(bool), min_size=min_size)

    nib.save(nib.Nifti1Image(img_out.astype(np.int16), img.affine), output_path)
    print(f"INFO - Saving generated image at\n\t{output_path :}")
    return output_path

def tissue_classifier(input_path, pve=True, force=True, suffix="fast"):
    """ Normalize image between 0 and 1.

    Parameters
    ----------
    input_data : str
        The image path that will be normalized

    suffix : str
        Nifti file suffixe for the new generated image
    """
    print(f"INFO - Starting tissue classification for\n\t{input_path :}")
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    pve_path = output_path.replace(".nii.gz", "_pve.nii.gz")

    if os.path.exists(output_path) and not force:
        print(f"INFO - Tissue classification already done for\n\t{input_path :}")
        if pve == True:
            return output_path, pve_path
        else:
            return output_path

    nclass = 3
    beta = 0.1

    img = nib.load(input_path)
    img_data = img.get_fdata()

    hmrf = TissueClassifierHMRF(verbose=False)
    _, img_out, mask_out = hmrf.classify(img_data, nclass, beta)

    print(f"INFO - Saving generated image at\n\t{output_path :}")
    nib.save(nib.Nifti1Image(img_out, img.affine), output_path)
    if pve == True:
        print(f"\t{pve_path :}")
        nib.save(nib.Nifti1Image(mask_out, img.affine), pve_path)
        return output_path, pve_path
    else:
        return output_path

def add_tissue_class(input_path, mask_path, num_class, force=True, suffix="masked"):
    print(f"INFO - Starting to add a new class for\n\t{input_path :}")
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - New added class already done for\n\t{input_path :}")
        return output_path
    
    img = nib.load(input_path)
    img_data = img.get_fdata()

    mask = nib.load(mask_path)
    mask_data = mask.get_fdata()

    img_data[mask_data > 0] = num_class

    nib.save(nib.Nifti1Image(img_data, img.affine), output_path)
    print(f"INFO - Saving generated image at\n\t{output_path :}")
    return output_path

def extract_dim(input_path, dim, force=True, suffix=""):
    print(f"INFO - Starting to extract the dimension {dim} for\n\t{input_path :}")
    output_path = input_path.replace(".nii.gz", "_dim" + str(dim) + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Extracted dimension already done for\n\t{input_path :}")
        return output_path
    
    img = nib.load(input_path)
    img_data = img.get_fdata()

    img_data = img_data[..., dim]
 
    nib.save(nib.Nifti1Image(img_data, img.affine), output_path)
    print(f"INFO - Saving generated image at\n\t{output_path :}")
    return output_path

def prediction_glioma(input_path, model_path, landmark_path, force=True, suffix="glioma_predicted"):
    print(f"INFO - Starting glioma prediction for\n\t{input_path :}")
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Glioma segmentation already done for\n\t{input_path :}")
        return output_path
    
    from monai.inferers import sliding_window_inference
    from monai.networks.nets import UNETR
    model = UNETR(
            in_channels=1,
            out_channels=2,
            img_size=(176, 208, 160),
            feature_size=16,
            hidden_size=768,
            mlp_dim=3072,
            num_heads=12,
            pos_embed="perceptron",
            norm_name="instance",
            res_block=True,
            dropout_rate=0.2,
    ).to(torch.device("cuda" if torch.cuda.is_available() else "cpu"))
    model.load_state_dict(torch.load(model_path))
    model.eval()

    transform = tio.Compose([
            tio.ToCanonical(),
            tio.Resample(1),
            tio.CropOrPad((176, 208, 160)),
            tio.HistogramStandardization({"image": np.load(landmark_path)}),
            tio.ZNormalization(masking_method=tio.ZNormalization.mean),
    ])

    ds = tio.SubjectsDataset([
        tio.Subject(image = tio.ScalarImage(input_path))], 
        transform=transform
    )[0]

    affine = nib.load(input_path).affine
    with torch.no_grad():
        img = ds["image"]["data"]
        val_inputs = torch.unsqueeze(img, 1)
        val_outputs = sliding_window_inference(val_inputs.cuda(), (176, 208, 160), 4, model, overlap=0.25)
        val_outputs = torch.argmax(val_outputs, dim=1).detach().cpu().numpy()[0].astype(float)
        pred_map = tio.LabelMap(tensor=np.expand_dims(val_outputs, 0), affine=ds.image.affine)
        ds.add_image(pred_map, "pred")
        ds_inv = ds.apply_inverse_transform(warn=True)
        val_outputs = ds_inv["pred"].data.numpy().squeeze()

        nib.save(nib.Nifti1Image(val_outputs.squeeze(), affine), output_path)
    
    output_path = remove_small_object(output_path, 5000, force=True)
    print(f"INFO - Saving generated image at\n\t{output_path :}")
    return output_path

def prediction_multiple_sclerosis(input_path, model_path, landmark_path, force=True, suffix="ms_predicted"):
    print(f"INFO - Starting multiple sclerosis prediction for\n\t{input_path :}")
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Multiple sclerosis segmentation already done for\n\t{input_path :}")
        return output_path
    
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    model = MPUnetPP2CBND(1, 1).to(device)
    model.load_state_dict(torch.load(model_path))
    model.eval()

    transform = tio.Compose([
        tio.ToCanonical(),
        tio.Resample(1),
        # tio.CopyAffine('image'),
        tio.CropOrPad((192, 192, 224)),
        tio.HistogramStandardization({"image": landmark_path}),
        tio.ZNormalization(masking_method=tio.ZNormalization.mean),
    ])

    subject = tio.SubjectsDataset([ tio.Subject( image = tio.ScalarImage(input_path) )],  transform=transform )[0]

    patch_overlap = 16
    grid_sampler = tio.inference.GridSampler(
        subject,
        64,
        patch_overlap
    )

    patch_loader = torch.utils.data.DataLoader(grid_sampler, batch_size=16)
    aggregator = tio.inference.GridAggregator(grid_sampler, overlap_mode="average")

    with torch.no_grad():
        for patches_batch in patch_loader:
            inputs = patches_batch['image'][tio.DATA].to(device)
            locations = patches_batch[tio.LOCATION]
            batch_pred = model(inputs)
            aggregator.add_batch(batch_pred, locations)
    output = aggregator.get_output_tensor()
    y_pred = output[0].numpy().reshape(subject.image.shape)

    val_inputs  = subject.image.numpy().squeeze()
    # val_outputs = torch.argmax(val_outputs, dim=1).detach().cpu().numpy()[0].astype(float)
    val_outputs = y_pred.squeeze()

    affine = nib.load(input_path).affine

    output_path = input_path.replace(".nii", "_predicted.nii")
    input_path = input_path.replace(".nii", "_input.nii")
    nib.save(
        nib.Nifti1Image(val_inputs, affine), 
        input_path
    )

    val_outputs[val_outputs <= 0.01] = 0
    val_outputs[val_outputs > 0.01] = 1
    nib.save(
        nib.Nifti1Image(val_outputs, affine), 
        output_path
    )

    print(f"INFO - Saving generated image at\n\t{output_path :}")
    return output_path, input_path

def uncertainty_prediction_glioma(input_path, model_path, force=True, suffix="uncertainty"):
    print(f"INFO - Starting glioma uncertainty prediction for\n\t{input_path :}")
    output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Uncertainty glioma segmentation already done for\n\t{input_path :}")
        return output_path
    
    from monai.inferers import sliding_window_inference
    from monai.networks.nets import UNETR
    from tqdm.auto import tqdm, trange
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model = UNETR(
            in_channels=1,
            out_channels=2,
            img_size=(176, 208, 160),
            feature_size=16,
            hidden_size=768,
            mlp_dim=3072,
            num_heads=12,
            pos_embed="perceptron",
            norm_name="instance",
            res_block=True,
            dropout_rate=0.2,
    ).to(device)
    model.load_state_dict(torch.load(model_path))
    model.eval()

    transform = tio.Compose([
            tio.ToCanonical(),
            tio.Resample(1),
            tio.CropOrPad((176, 208, 160)),
            tio.HistogramStandardization({"image": np.load("E:/Leo/script/results/landmarks.npy")}),
            tio.ZNormalization(masking_method=tio.ZNormalization.mean),
            tio.RandomFlip(),
            tio.RandomAffine(p=0.5),
    ])

    subject = tio.Subject(image = tio.ScalarImage(input_path)) 
    affine = nib.load(input_path).affine

    results = []
    for _ in trange(20):
        subject = transform(subject)
        inputs = subject.image.data.to(device)
        
        with torch.no_grad():
            inputs = torch.unsqueeze(inputs, 1)
            outputs = sliding_window_inference(inputs, (176, 208, 160), 4, model, overlap=0.25)
        outputs = torch.argmax(outputs, dim=1).detach().cpu().numpy()[0].astype(float)
        pred_map = tio.LabelMap(tensor=np.expand_dims(outputs, 0), affine=subject.image.affine)
        subject.add_image(pred_map, "pred")
        subject_inv = subject.apply_inverse_transform(warn=True)
        results.append(subject_inv["pred"].data)

    result = torch.stack(results).long()
    tta_result_tensor = result.mode(dim=0).values

    different = torch.stack([
        tensor != tta_result_tensor
        for tensor in results
    ])
    uncertainty = different.float().mean(dim=0)
    uncertainty_img = tio.ScalarImage(tensor=uncertainty, affine=subject.image.affine)
    subject.add_image(uncertainty_img, "uncertainty")

    uncertainty_img = uncertainty_img.data.numpy().squeeze()
    nib.save(nib.Nifti1Image(uncertainty_img, affine), output_path)

    print(f"INFO - Saving generated image at\n\t{output_path :}")
    return output_path

    print(f"INFO - Starting glioma prediction for\n\t{input_path :}")
    
    from monai.inferers import sliding_window_inference
    from monai.networks.nets import UNETR
    model = UNETR(
            in_channels=1,
            out_channels=2,
            img_size=(176, 208, 160),
            feature_size=16,
            hidden_size=768,
            mlp_dim=3072,
            num_heads=12,
            pos_embed="perceptron",
            norm_name="instance",
            res_block=True,
            dropout_rate=0.2,
    ).to(torch.device("cuda" if torch.cuda.is_available() else "cpu"))
    model.load_state_dict(torch.load(model_path))
    model.eval()

    transform = tio.Compose([
            tio.ToCanonical(),
            tio.Resample(1),
            tio.CopyAffine('image'),
            tio.CropOrPad((176, 208, 160)),
            tio.ZNormalization(masking_method=tio.ZNormalization.mean),
    ])

    ds = tio.SubjectsDataset([
        tio.Subject(image = tio.ScalarImage(input_path))], 
        transform=transform
    )[0]

    affine = nib.load(input_path).affine
    with torch.no_grad():
        img = ds["image"]["data"]
        val_inputs = torch.unsqueeze(img, 1)
        val_outputs = sliding_window_inference(val_inputs.cuda(), (176, 208, 160), 4, model, overlap=0.25)

        pred_map = tio.LabelMap(tensor=val_outputs[0], affine=ds.image.affine)
        ds.add_image(pred_map, "pred")
        val_outputs = ds.apply_inverse_transform(warn=True).pred.data

        val_outputs = torch.argmax(val_outputs, dim=1).detach().cpu().numpy()[0].astype(float)
        val_outputs_path = input_path.replace(".nii", "_predicted_unetr.nii")
        nib.save(nib.Nifti1Image(val_outputs, affine), val_outputs_path)
    
    print(f"INFO - Saving generated image at\n\t{val_outputs_path :}")
    return val_outputs_path

if __name__ == "__main__":
    landmark_path = r"E:\SEP_IA\results_7T\final\landmarks.npy"
    model_path = r"E:\SEP_IA\results_7T\final\patches_epoch_298.pth"
    input_path = r"E:\SEP_IA\data\sub-001\ses-01\anat\sub-001_ses-01_FLAIR.nii.gz"

    prediction_multiple_sclerosis(input_path, model_path, landmark_path)


import os

from .utils import *
from .transformation import *

global MODEL_PATH, FORCE

MODEL_PATH = None
FORCE = False

def init(model_path=None, force=False):
    global MODEL_PATH, FORCE
    if model_path is not None and not os.path.isdir(model_path):
        raise ValueError(f"ERROR - Could not find model folds in the following directory: {model_path}")
    else:
        MODEL_PATH = model_path
    FORCE = force

def get_model():
    global MODEL_PATH
    return MODEL_PATH

def get_force():
    global FORCE
    return FORCE

@timed
def preproc(sub_path, ses="ses-01", ref="T1w", normalize=False, keep_all=True):
    # Checking errors
    sub = os.path.basename(sub_path)
    if "sub" not in sub:
        raise ValueError("ERROR - Could not find a sub number in the data_path, make sure it is bids compliant")

    ses_path = os.path.join(sub_path, ses)
    if not os.path.isdir(ses_path):
        raise ValueError(f"ERROR - Could not find a directory with the following session number {ses}")

    modalities = os.listdir(ses_path)
    if "anat" not in modalities:
        raise ValueError("ERROR - Can not start process without the anat directory")

    anat_path = os.path.join(ses_path, "anat")
    ref_path = f"{sub}_{ses}_{ref}.nii.gz"
    if ref_path not in os.listdir(anat_path):
        raise FileNotFoundError(f"ERROR - The following reference filename could not be found:\n\t{ref_path}")

    # Starting to preproc the reference sequence
    print(f"INFO - Starting preprocessing for the reference image at\n\t{ref_path}")
    ref_path = os.path.join(anat_path, ref_path)
    ref_corrected, ref_brain_mask, ref_crop = ref_preproc(ref_path, normalize)

    # Starting to loop through each sequence
    print(f"INFO - Starting preprocessing for the following modalities:\n\t{', '.join(modalities)}")
    print("INFO - Starting with anatomic sequences")
    for seq in os.listdir(anat_path):
        seq_path = os.path.join(anat_path, seq)
        # Check if the file is a nii.gz
        if seq_path.endswith("nii.gz") and seq_path != ref_path and is_native(seq_path):
            print(f"INFO - Nifti file found: {seq_path}")
            # Starting preproc for the current sequence path
            seq_path = other_preproc(seq_path, ref_corrected, ref_brain_mask, ref_crop, normalize)

    print("INFO - Continuing with diffusion sequences")
    dwi_path = os.path.join(ses_path, "dwi")
    for seq in os.listdir(dwi_path):
        seq_path = os.path.join(dwi_path, seq)
        # Check if the file is a nii.gz
        if seq_path.endswith("nii.gz") and is_native(seq_path) and is_useful(seq_path):
            print(f"INFO - Nifti file found: {seq_path}")
            # Starting preproc for the current sequence path
            seq_path = other_preproc(seq_path, ref_corrected, ref_brain_mask, ref_crop, normalize)

    print("INFO - Continuing with perfusion sequences")
    perf_path = os.path.join(ses_path, "perf")
    for seq in os.listdir(perf_path):
        seq_path = os.path.join(perf_path, seq)
        # Check if the file is a nii.gz
        if seq_path.endswith("nii.gz") and is_native(seq_path) and is_useful(seq_path):
            print(f"INFO - Nifti file found: {seq_path}")
            # Starting preproc for the current sequence path
            seq_path = other_preproc(seq_path, ref_corrected, ref_brain_mask, ref_crop, normalize)

@timed
def ref_preproc(ref_path, normalize):
    # crop => resample => n4 bias field correction => skull stripping => crop
    ref_path_cropped, crop_idx_1 = crop(ref_path, force=get_force())
    ref_path_resampled = resample(ref_path_cropped, 1)[0]
    ref_path_corrected, ref_path_corrected_mask = n4_bias_field_correction(ref_path_resampled, mask=True, force=get_force())
    ref_path_brain, ref_path_brain_mask = skull_stripping(ref_path_corrected, get_model(), mask=True, force=get_force())
    ref_path_brain_cropped, crop_idx_2 = crop(ref_path_brain, force=get_force())
    ref_path_brain_mask_cropped = apply_crop(ref_path_brain_mask, crop_idx_2, force=get_force())
    if normalize: ref_path_normalized = normalize(ref_path_brain_cropped, force=get_force())
    return ref_path_corrected, ref_path_brain_mask, crop_idx_2

@timed
def other_preproc(seq_path, ref_brain, ref_brain_mask, ref_crop, normalize):
    # registration => apply brain mask => apply crop => n4 bias field correction
    seq_path_registered, matrix_path = registration(ref_brain, seq_path, force=get_force())
    seq_path_brain = apply_mask(seq_path_registered, ref_brain_mask, suffix="brain", force=get_force())
    seq_path_cropped = apply_crop(seq_path_brain, crop=ref_crop, force=get_force())
    seq_path_corrected, seq_path_corrected_mask = n4_bias_field_correction(seq_path_cropped, mask=True, force=get_force())
    if normalize: seq_path_normalized = normalize(seq_path_corrected, force=get_force())

# def prediction_glioma(input_path, model_path, landmarks_path, force=True, suffix="predicted"):
#     print(f"INFO - Starting glioma prediction for\n\t{input_path :}")
#     output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
#     if os.path.exists(output_path) and not force:
#         print(f"INFO - Glioma segmentation already done for\n\t{input_path :}")
#         return output_path
    
#     from monai.inferers import sliding_window_inference
#     from monai.networks.nets import UNETR
#     import torch
#     import torchio as tio
#     import nibabel as nib
#     import numpy as np
#     model = UNETR(
#             in_channels=1,
#             out_channels=2,
#             img_size=(176, 208, 160),
#             feature_size=16,
#             hidden_size=768,
#             mlp_dim=3072,
#             num_heads=12,
#             pos_embed="perceptron",
#             norm_name="instance",
#             res_block=True,
#             dropout_rate=0.2,
#     ).to(torch.device("cuda" if torch.cuda.is_available() else "cpu"))
#     model.load_state_dict(torch.load(model_path))
#     model.eval()

#     transform = tio.Compose([
#             tio.ToCanonical(),
#             tio.Resample(1),
#             tio.CropOrPad((176, 208, 160)),
#             tio.HistogramStandardization({"image": np.load(landmarks_path)}),
#             tio.ZNormalization(masking_method=tio.ZNormalization.mean),
#     ])

#     ds = tio.SubjectsDataset([
#         tio.Subject(image = tio.ScalarImage(input_path))], 
#         transform=transform
#     )[0]

#     affine = nib.load(input_path).affine
#     with torch.no_grad():
#         img = ds["image"]["data"]
#         val_inputs = torch.unsqueeze(img, 1)
#         val_outputs = sliding_window_inference(val_inputs.cuda(), (176, 208, 160), 4, model, overlap=0.25)
#         val_outputs = torch.argmax(val_outputs, dim=1).detach().cpu().numpy()[0].astype(float)
#         pred_map = tio.LabelMap(tensor=np.expand_dims(val_outputs, 0), affine=ds.image.affine)
#         ds.add_image(pred_map, "pred")
#         ds_inv = ds.apply_inverse_transform(warn=True)
#         val_outputs = ds_inv["pred"].data.numpy().squeeze()

#         nib.save(nib.Nifti1Image(val_outputs.squeeze(), affine), output_path)
    
#     output_path = remove_small_object(output_path, 5000, force=True)
#     print(f"INFO - Saving generated image at\n\t{output_path :}")
#     return output_path

# def uncertainty_prediction_glioma(input_path, model_path, force=True, suffix="uncertainty"):
#     print(f"INFO - Starting glioma uncertainty prediction for\n\t{input_path :}")
#     output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
#     if os.path.exists(output_path) and not force:
#         print(f"INFO - Uncertainty glioma segmentation already done for\n\t{input_path :}")
#         return output_path
    
#     from monai.inferers import sliding_window_inference
#     from monai.networks.nets import UNETR
#     import torch
#     import torchio as tio
#     import nibabel as nib
#     import numpy as np
#     from tqdm.auto import tqdm, trange
#     device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
#     model = UNETR(
#             in_channels=1,
#             out_channels=2,
#             img_size=(176, 208, 160),
#             feature_size=16,
#             hidden_size=768,
#             mlp_dim=3072,
#             num_heads=12,
#             pos_embed="perceptron",
#             norm_name="instance",
#             res_block=True,
#             dropout_rate=0.2,
#     ).to(device)
#     model.load_state_dict(torch.load(model_path))
#     model.eval()

#     transform = tio.Compose([
#             tio.ToCanonical(),
#             tio.Resample(1),
#             tio.CropOrPad((176, 208, 160)),
#             tio.HistogramStandardization({"image": np.load("E:/Leo/script/results/landmarks.npy")}),
#             tio.ZNormalization(masking_method=tio.ZNormalization.mean),
#             tio.RandomFlip(),
#             tio.RandomAffine(p=0.5),
#     ])

#     subject = tio.Subject(image = tio.ScalarImage(input_path)) 
#     affine = nib.load(input_path).affine

#     results = []
#     for _ in trange(20):
#         subject = transform(subject)
#         inputs = subject.image.data.to(device)
        
#         with torch.no_grad():
#             inputs = torch.unsqueeze(inputs, 1)
#             outputs = sliding_window_inference(inputs, (176, 208, 160), 4, model, overlap=0.25)
#         outputs = torch.argmax(outputs, dim=1).detach().cpu().numpy()[0].astype(float)
#         pred_map = tio.LabelMap(tensor=np.expand_dims(outputs, 0), affine=subject.image.affine)
#         subject.add_image(pred_map, "pred")
#         subject_inv = subject.apply_inverse_transform(warn=True)
#         results.append(subject_inv["pred"].data)

#     result = torch.stack(results).long()
#     tta_result_tensor = result.mode(dim=0).values

#     different = torch.stack([
#         tensor != tta_result_tensor
#         for tensor in results
#     ])
#     uncertainty = different.float().mean(dim=0)
#     uncertainty_img = tio.ScalarImage(tensor=uncertainty, affine=subject.image.affine)
#     subject.add_image(uncertainty_img, "uncertainty")

#     uncertainty_img = uncertainty_img.data.numpy().squeeze()
#     nib.save(nib.Nifti1Image(uncertainty_img, affine), output_path)

#     print(f"INFO - Saving generated image at\n\t{output_path :}")
#     return output_path

# def add_tissue_class(input_path, mask_path, num_class, force=True, suffix="masked"):
#     import nibabel as nib

#     print(f"INFO - Starting to add a new class for\n\t{input_path :}")
#     output_path = input_path.replace(".nii.gz", "_" + suffix + ".nii.gz")
#     if os.path.exists(output_path) and not force:
#         print(f"INFO - New added class already done for\n\t{input_path :}")
#         return output_path
    
#     img = nib.load(input_path)
#     img_data = img.get_fdata()

#     mask = nib.load(mask_path)
#     mask_data = mask.get_fdata()

#     img_data[mask_data > 0] = num_class

#     nib.save(nib.Nifti1Image(img_data, img.affine), output_path)
#     print(f"INFO - Saving generated image at\n\t{output_path :}")
#     return output_path

# def extract_dim(input_path, dim, force=True, suffix=""):
#     import nibabel as nib

#     print(f"INFO - Starting to extract the dimension {dim} for\n\t{input_path :}")
#     output_path = input_path.replace(".nii.gz", "_dim" + str(dim) + ".nii.gz")
#     if os.path.exists(output_path) and not force:
#         print(f"INFO - Extracted dimension already done for\n\t{input_path :}")
#         return output_path
    
#     img = nib.load(input_path)
#     img_data = img.get_fdata()

#     img_data = img_data[..., dim]
 
#     nib.save(nib.Nifti1Image(img_data, img.affine), output_path)
#     print(f"INFO - Saving generated image at\n\t{output_path :}")
#     return output_path

class Pipeline:
    def __init__(self, data):
        self.data = data
        self.clean()

        if not os.path.exists(self.data): 
            raise FileNotFoundError(f"ERROR - The following data file is not existing: {self.data}")

    def clean(self):
        self.cropped_idx = None
        self.matrix = None
        self.bias = None
        self.mask = None
        self.pred = None
        self.pve = None

    def susan(self, size):
        self.data = susan(self.data, size)

    def remove_small_object(self, threshold):
        self.data = remove_small_object(self.data, threshold)

    def tissue_classifier(self):
        self.data, self.pve = tissue_classifier(self.data, force=False)

    def skull_stripping(self, mask, model):
        if not os.path.isdir(model): 
            raise NotADirectoryError(f"ERROR - The following model path is not a directory: {model}")
        
        if mask:
            self.data, self.mask = skull_stripping(self.data, model, mask, force=False)
        else:
            self.data = skull_stripping(self.data, model, mask, force=False)

    def registration(self, ref):
        self.data, self.matrix = registration(ref, self.data, force=False)

    def apply_transformation(self, ref, matrix):
        if not os.path.exists(ref): 
            raise FileNotFoundError(f"ERROR - The following reference file is not existing: {ref}")
        if not os.path.exists(matrix): 
            raise FileNotFoundError(f"ERROR - The following matrix file is not existing: {matrix}")
        self.data = apply_transformation(ref, self.data, matrix)

    def crop(self):
        self.data, self.cropped_idx = crop(self.data)
        print(f"DEBUG - Cropping at {self.cropped_idx[0]}, {self.cropped_idx[1]}, {self.cropped_idx[2]}, {self.cropped_idx[3]}, {self.cropped_idx[4]}, {self.cropped_idx[5]}")

    def resample(self, obj):
        self.data = resample(self.data, obj)[0]

    def apply_crop(self, crop=None):
        if self.cropped_idx is None:
            if crop is not None:
                self.data = apply_crop(self.data, crop)
            else:
                raise ValueError("ERROR - If no previous crop done, you need to have a crop argument")
        else:
            self.data = apply_crop(self.data, self.cropped_idx)

    def apply_mask(self, mask):
        self.data = apply_mask(self.data, mask)

    def normalize(self):
        self.data = normalize(self.data)

    def n4_bias_field_correction(self, mask):
        if mask:
            self.data, self.bias = n4_bias_field_correction(self.data, mask, force=False)
        else:
            self.data = n4_bias_field_correction(self.data, mask, force=False)

    def prediction_glioma(self, model, landmarks):
        self.pred = prediction_glioma(self.data, model, landmarks)

    def add_tissue_class(self, ref, num):
        self.data = add_tissue_class(ref, self.pred, num)

    def susan(self):
        self.data = susan(self.data, force=False)

    def extract_dim(self, dim):
        self.data = extract_dim(self.data, dim)

    def change_data(self, new_data):
        self.data = new_data

if __name__ == '__main__':
    sub = "sub-031"
 
    # Correction IA
    """
    pipeline = Pipeline(f"E:/Haissam/data/{sub}/ses-01/anat/{sub}_ses-01_T1w.nii.gz")
    pipeline.pred = f"E:/Haissam/data/{sub}/ses-01/anat/{sub}_ses-01_FLAIR_flirt_mask_susan_predicted_filtered.nii.gz"
    tissues = f"E:/Haissam/data/{sub}/ses-01/anat/{sub}_ses-01_T1w_cropped_resampled_corrected_brain_fast.nii.gz"
    pipeline.add_tissue_class(tissues, num=4)
    raise
    """

    # T1
    pipeline = Pipeline(f"E:/Haissam/data/{sub}/ses-01/anat/{sub}_ses-01_T1w.nii.gz")
    pipeline.crop()
    pipeline.resample(1)
    pipeline.n4_bias_field_correction(mask=True)
    pipeline.skull_stripping(mask=True, model=f"C:/Users/467355/Documents/HD-BET-master/HD_BET/hd-bet_params")
    path = pipeline.data
    pipeline.tissue_classifier()
    tissues = pipeline.data
    # T2 FLAIR
    pipeline.change_data(f"E:/Haissam/data/{sub}/ses-01/anat/{sub}_ses-01_FLAIR.nii.gz")
    pipeline.registration(ref=path)
    pipeline.apply_mask(mask=pipeline.mask)
    pipeline.susan()
    pipeline.prediction_glioma(model="E:/Leo/script/results/new_finetuning_susan_aug/not_best_metric_model_75000.pth", landmarks="E:/Leo/script/results/landmarks.npy")
    pipeline.add_tissue_class(tissues, num=4)
    # T1ce
    pipeline.change_data(f"E:/Haissam/data/{sub}/ses-01/anat/{sub}_ses-01_ce-GADOLINIUM_T1w.nii.gz")
    pipeline.registration(ref=path)
    pipeline.apply_mask(mask=pipeline.mask)
    # SWI
    pipeline.change_data(f"E:/Haissam/data/{sub}/ses-01/anat/{sub}_ses-01_SWI.nii.gz")
    pipeline.registration(ref=path)
    pipeline.apply_mask(mask=pipeline.mask)
    # Resolve ADC
    pipeline.change_data(f"E:/Haissam/data/{sub}/ses-01/dwi/{sub}_ses-01_resolve_adc.nii.gz")
    pipeline.registration(ref=path)
    pipeline.apply_mask(mask=pipeline.mask)
    # Resolve Tracew
    pipeline.change_data(f"E:/Haissam/data/{sub}/ses-01/dwi/{sub}_ses-01_resolve_tracew.nii.gz")
    pipeline.extract_dim(1)
    pipeline.registration(ref=path)
    pipeline.apply_mask(mask=pipeline.mask)
    # FA
    pipeline.change_data(f"E:/Haissam/data/{sub}/ses-01/dwi/{sub}_ses-01_fa.nii.gz")
    pipeline.registration(ref=path)
    pipeline.apply_mask(mask=pipeline.mask)
    # T2* DSC PERFUSION
    pipeline.change_data(f"E:/Haissam/data/{sub}/ses-01/perf/{sub}_ses-01_perf.nii.gz")
    pipeline.extract_dim(0)
    pipeline.registration(ref=path)
    pipeline.apply_mask(mask=pipeline.mask)
    # CBV
    pipeline.change_data(f"E:/Haissam/data/{sub}/ses-01/perf/{sub}_ses-01_rCBV.nii.gz")
    pipeline.apply_transformation(ref=path, matrix=pipeline.matrix)
    pipeline.apply_mask(mask=pipeline.mask)
    # CBF
    pipeline.change_data(f"E:/Haissam/data/{sub}/ses-01/perf/{sub}_ses-01_rCBF.nii.gz")
    pipeline.apply_transformation(ref=path, matrix=pipeline.matrix)
    pipeline.apply_mask(mask=pipeline.mask)
    # MTT
    pipeline.change_data(f"E:/Haissam/data/{sub}/ses-01/perf/{sub}_ses-01_MTT.nii.gz")
    pipeline.apply_transformation(ref=path, matrix=pipeline.matrix)
    pipeline.apply_mask(mask=pipeline.mask)
    # TPP
    pipeline.change_data(f"E:/Haissam/data/{sub}/ses-01/perf/{sub}_ses-01_TTP.nii.gz")
    pipeline.apply_transformation(ref=path, matrix=pipeline.matrix)
    pipeline.apply_mask(mask=pipeline.mask)



import os
import nibabel as nib
import numpy as np
import pydicom
import math

from .utils import load_dicom

def generate_mask(svs_path, nii_path, force=True, output_suffixe="mask"):
    """ Create a 3D mask of the spectroscopy volume of interest

    Parameters
    ----------
    svs_path : str 
        Path of the spectroscopy DICOM file
    
    nii_path : str
        Path of the reference volume

    output_suffixe : str
        Nifti file suffixe for the new generated image
    """
    output_path = nii_path.replace(".nii.gz", "_" + output_suffixe + ".nii.gz")
    if os.path.exists(output_path) and not force:
        print(f"INFO - Mask already generated for\n\t{nii_path :}")
        return output_path

    ref = nib.load(nii_path)
    ref_array = ref.get_fdata()

    pixdim = ref.header["pixdim"][1:4]

    csa_header = load_dicom(svs_path)

    voi = np.array([
        round(float(csa_header["sSpecPara.sVoI.dReadoutFOV"]) / pixdim[0]),
        round(float(csa_header["sSpecPara.sVoI.dPhaseFOV"])   / pixdim[1]), 
        round(float(csa_header["sSpecPara.sVoI.dThickness"])  / pixdim[2])
    ], dtype=np.int8)

    pcg_centre = [
        float(csa_header["sSpecPara.sVoI.sPosition.dSag"]),
        float(csa_header["sSpecPara.sVoI.sPosition.dCor"]), 
        float(csa_header["sSpecPara.sVoI.sPosition.dTra"])
    ]
    pcg_centre.append(1)
    pcg_centre_index = np.matmul(np.linalg.inv(ref.affine), pcg_centre).squeeze()

    mask = np.zeros((ref_array.shape))
    mask[
        round(pcg_centre_index[0] - voi[0]/2) : round(pcg_centre_index[0] + voi[0]/2),
        round(pcg_centre_index[1] - voi[1]/2) : round(pcg_centre_index[1] + voi[1]/2),
        round(pcg_centre_index[2] - voi[2]/2) : round(pcg_centre_index[2] + voi[2]/2),
    ] = 1

    mask = np.ma.masked_where(mask==1, mask)
    mask = mask[::-1, ::-1, :]
    mask_nii = nib.Nifti1Image(mask, ref.affine)
    print(f"INFO - Saving generated image at\n\t{output_path :}")
    nib.save(mask_nii, output_path)
    return output_path

def generate_temporal_spectre(svs_path, spectre_path, save=True):
    dcm = pydicom.dcmread(svs_path)
    spectro_data = dcm[0x5600,0x0020].value
    spectre = np.fromstring(spectro_data, np.float32)
    np.save(spectre_path, spectre)
    if save:
        return spectre_path
    else:
        return spectre
    
def generate_frequency_spectre(path):
    if os.path.endswith(".dcm"):
        temporal_spectre = generate_temporal_spectre(path, save=False) 
    else:
        temporal_spectre = np.load(path)

    shape = int(temporal_spectre.shape[0]/2)
    complex_spectre = np.zeros((shape,), dtype=complex)
    cpt = 0 
    for i in range(shape):
        compl = temporal_spectre[cpt] + 1j * temporal_spectre[cpt+1]
        complex_spectre[i] = np.array(compl, dtype=complex)
        cpt += 2

    frequency_spectre = np.fft.fftshift(np.fft.fft(complex_spectre))
    np.save(path.replace(".npy", "_fft.npy").replace(".dcm", "_fft.npy"), frequency_spectre)
    return frequency_spectre